#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#
# pylint: disable=invalid-name

"""Dataloader of the Car Connection Picture Dataset"""

from .CarConnection import CarConnection

__all__ = ["CarConnection"]
